import click
from gefyra.cli.console import info


@click.command()
@click.option(
    "-n",
    "--no-check",
    help="Do not check whether there is a new version",
    is_flag=True,
    default=False,
)
@click.pass_context
def version(ctx, no_check):
    import requests
    import gefyra.configuration as config

    info(f"Gefyra client version: {config.__VERSION__}")
    if not no_check:
        release = requests.get(
            "https://api.github.com/repos/gefyrahq/gefyra/releases/latest"
        )
        if release.status_code == 403:  # pragma: no cover
            info("Versions cannot be compared, as API rate limit was exceeded")
            return None
        latest_release_version = release.json()["tag_name"].replace("-", ".")
        if config.__VERSION__ != latest_release_version:  # pragma: no cover
            info(
                f"You are using gefyra version {config.__VERSION__}; however, version"
                f" {latest_release_version} is available."
            )
    return True



# flake8: noqa
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from gefyra.types import GefyraInstallOptions


def data(params: "GefyraInstallOptions") -> list[dict]:
    return [
        {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": {
                "name": "gefyra-operator",
                "namespace": params.namespace,
                "labels": {
                    "gefyra.dev/app": "gefyra-operator",
                    "gefyra.dev/role": "operator",
                },
            },
            "spec": {
                "replicas": 1,
                "selector": {
                    "matchLabels": {
                        "gefyra.dev/app": "gefyra-operator",
                        "gefyra.dev/role": "operator",
                    }
                },
                "template": {
                    "metadata": {
                        "labels": {
                            "gefyra.dev/app": "gefyra-operator",
                            "gefyra.dev/role": "operator",
                        }
                    },
                    "spec": {
                        "serviceAccountName": "gefyra-operator",
                        "containers": [
                            {
                                "name": "gefyra",
                                "image": f"quay.io/gefyra/operator:{params.version}",
                                "imagePullPolicy": "Always",
                                "ports": [{"containerPort": 9443}],
                                "env": [
                                    {
                                        "name": "GEFYRA_STOWAWAY_TAG",
                                        "value": params.version,
                                    },
                                    {
                                        "name": "GEFYRA_CARRIER_IMAGE_TAG",
                                        "value": params.version,
                                    },
                                ],
                            }
                        ],
                    },
                },
            },
        },
        {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": {
                "name": "gefyra-operator-webhook",
                "namespace": params.namespace,
                "labels": {
                    "gefyra.dev/app": "gefyra-operator",
                    "gefyra.dev/role": "webhook",
                },
            },
            "spec": {
                "replicas": 1,
                "selector": {
                    "matchLabels": {
                        "gefyra.dev/app": "gefyra-operator",
                        "gefyra.dev/role": "webhook",
                    }
                },
                "template": {
                    "metadata": {
                        "labels": {
                            "gefyra.dev/app": "gefyra-operator",
                            "gefyra.dev/role": "webhook",
                        }
                    },
                    "spec": {
                        "serviceAccountName": "gefyra-operator",
                        "containers": [
                            {
                                "name": "gefyra",
                                "image": f"quay.io/gefyra/operator:{params.version}",
                                "imagePullPolicy": "Always",
                                "ports": [{"containerPort": 9443}],
                                "env": [
                                    {"name": "OP_MODE", "value": "webhook"},
                                    {
                                        "name": "GEFYRA_STOWAWAY_TAG",
                                        "value": params.version,
                                    },
                                    {
                                        "name": "GEFYRA_CARRIER_IMAGE_TAG",
                                        "value": params.version,
                                    },
                                ],
                                "livenessProbe": {
                                    "exec": {
                                        "command": [
                                            "python",
                                            "gefyra/healthcheck.py",
                                        ]
                                    },
                                    "initialDelaySeconds": 5,
                                    "periodSeconds": 5,
                                },
                            }
                        ],
                    },
                },
            },
        },
    ]



import time

import logging
from gefyra.configuration import ClientConfiguration
from gefyra.exceptions import (
    GefyraClientAlreadyExists,
    GefyraClientNotFound,
    GefyraConnectionError,
)

logger = logging.getLogger(__name__)


def handle_create_gefyraclient(config: ClientConfiguration, body) -> dict:
    from kubernetes.client import ApiException

    retries = 15
    counter = 0
    success = False
    while not success:
        try:
            gclient = config.K8S_CUSTOM_OBJECT_API.create_namespaced_custom_object(
                namespace=config.NAMESPACE,
                body=body,
                group="gefyra.dev",
                plural="gefyraclients",
                version="v1",
            )
            success = True
        except ApiException as e:
            if e.status == 409:
                raise GefyraClientAlreadyExists(
                    f"Client {body['metadata']['name']} already exists."
                )
            logger.error(
                f"A Kubernetes API Error occured. \nReason:{e.reason} \nBody:{e.body}"
            )
            if counter < retries:
                counter += 1
                time.sleep(4)
            else:
                raise e
    return gclient


def handle_get_gefyraclient(config: ClientConfiguration, client_id: str) -> dict:
    from kubernetes.client import ApiException
    import urllib3

    try:
        gclient = config.K8S_CUSTOM_OBJECT_API.get_namespaced_custom_object(
            namespace=config.NAMESPACE,
            name=client_id,
            group="gefyra.dev",
            plural="gefyraclients",
            version="v1",
        )
    except ApiException as e:
        if e.status in [404, 403]:
            raise GefyraClientNotFound(f"Client {client_id} does not exist.")
        else:
            logger.error(
                f"A Kubernetes API Error occured. \nReason:{e.reason} \nBody:{e.body}"
            )
            raise e
    except urllib3.exceptions.MaxRetryError as e:
        # this connection does not work (at the moment)
        raise GefyraConnectionError(
            f"This connection does not work. Is the cluster at {e.pool.host}:{e.pool.port} reachable? "
            f"Is the client '{client_id}' stale (e.g. from an old connection)? "
            f"Remove it with 'gefyra connection remove {client_id}' and try again."
        )

    return gclient


def handle_delete_gefyraclient(
    config: ClientConfiguration, client_id: str, force: bool, wait: bool = False
) -> bool:
    from kubernetes.client import ApiException

    try:
        if force:
            config.K8S_CUSTOM_OBJECT_API.patch_namespaced_custom_object(
                namespace=config.NAMESPACE,
                name=client_id,
                group="gefyra.dev",
                plural="gefyraclients",
                version="v1",
                body={"metadata": {"finalizers": None}},
            )
        config.K8S_CUSTOM_OBJECT_API.delete_namespaced_custom_object(
            namespace=config.NAMESPACE,
            name=client_id,
            group="gefyra.dev",
            plural="gefyraclients",
            version="v1",
        )
        if wait:
            timeout = 30
            counter = 0
            while counter < timeout:
                try:
                    handle_get_gefyraclient(config=config, client_id=client_id)
                except GefyraClientNotFound:
                    return True
                time.sleep(1)
                counter += 1
            return False
        return True
    except ApiException as e:
        logger.debug(e)
        if e.status in [404, 403]:
            return False
        else:
            logger.error(
                f"A Kubernetes API Error occured. \nReason:{e.reason} \nBody:{e.body}"
            )
            raise e


def get_gefyraclient_body(
    config: ClientConfiguration, client_id: str, provider: str = "stowaway"
) -> dict:
    return {
        "apiVersion": "gefyra.dev/v1",
        "kind": "gefyraclient",
        "metadata": {
            "name": client_id,
            "namespace": config.NAMESPACE,
        },
        "provider": provider,
    }



import os
import click
from gefyra.cli import console
from gefyra.cli.utils import AliasedGroup, standard_error_handler
from tabulate import tabulate


@click.group(
    "clients", cls=AliasedGroup, help="Manage clients for a Gefyra installation"
)
@click.pass_context
def clients(ctx):
    # for management of clients we always sourcing the kubeconfig and context from env if not passed
    if ctx.obj["kubeconfig"] is None:
        ctx.obj["kubeconfig"] = os.environ.get("KUBECONFIG") or os.path.expanduser(
            "~/.kube/config"
        )


@clients.command("create", help="Create a new Gefyra client")
@click.option("--client-id", help="The client id", type=str)
@click.option(
    "-n",
    "quantity",
    help="Number of clients to be generated (not allowed with explicit client-id)",
    type=int,
    default=1,
)
@click.pass_context
@standard_error_handler
def create_clients(ctx, client_id, quantity):
    from gefyra import api

    api.add_clients(
        client_id,
        quantity,
        kubeconfig=ctx.obj["kubeconfig"],
        kubecontext=ctx.obj["context"],
    )
    console.success(f"{quantity} client(s) created successfully")


@clients.command(
    "delete", alias=["rm", "remove"], help="Mark a Gefyra client for deletion"
)
@click.argument("client_id", nargs=-1, required=True)
@click.pass_context
@standard_error_handler
def delete_client(ctx, client_id):
    from gefyra import api

    for _del in list(client_id):
        deleted = api.delete_client(
            _del, kubeconfig=ctx.obj["kubeconfig"], kubecontext=ctx.obj["context"]
        )
        if deleted:
            console.success(f"Client {_del} marked for deletion")


@clients.command("list", alias=["ls"], help="List all Gefyra clients")
@click.pass_context
@standard_error_handler
def list_client(ctx):
    from gefyra import api

    gefyraclients = api.list_client(
        kubeconfig=ctx.obj["kubeconfig"], kubecontext=ctx.obj["context"]
    )
    clients = [
        [
            c.client_id,
            c.state,
            c.state_transitions.get("CREATING", "Creating..."),
        ]
        for c in gefyraclients
    ]
    click.echo(tabulate(clients, headers=["ID", "STATE", "CREATED"], tablefmt="plain"))


@clients.command("inspect", alias=["show", "get"], help="Describe a Gefyra client")
@click.argument("client_id")
@click.pass_context
@standard_error_handler
def inspect_client(ctx, client_id):
    from gefyra import api

    client = api.get_client(
        client_id, kubeconfig=ctx.obj["kubeconfig"], kubecontext=ctx.obj["context"]
    )
    console.heading(client.client_id)
    console.info(f"uid: {client.uid}")
    console.info(f"States: {client.state_transitions}")


@clients.command(
    "config", alias=["write"], help="Get a Gefyra connection config for a client"
)
@click.argument("client_id")
@click.option("-h", "--host", help="The connection host", type=str)
@click.option("-p", "--port", help="The connection port (default: 31820)", type=int)
@click.option(
    "-a",
    "--kube-api",
    "--kubernetes-api",
    help=(
        "The Kubernetes API adress for the host cluster (default: API adresse of your"
        " kubeconfig)"
    ),
    type=str,
)
@click.option(
    "-o",
    "--output",
    help="The output file to write the config to",
    type=click.File("wb"),
)
@click.pass_context
@standard_error_handler
def get_config(ctx, client_id, host, port, kube_api, output):
    from gefyra import api

    json_str = api.write_client_file(
        client_id,
        host=host,
        port=port,
        kube_api=kube_api,
        kubeconfig=ctx.obj["kubeconfig"],
        kubecontext=ctx.obj["context"],
    )
    if output:
        output.write(json_str.encode("utf-8"))
    else:
        click.echo(json_str)



import click


def heading(text: str):
    click.echo(click.style(text, bold=True))


def error(text: str):
    click.echo(click.style(text, fg="red"))


def info(text: str):
    click.echo(text)


def success(text: str):
    click.echo(click.style(text, fg="green"))



import logging


from gefyra.configuration import ClientConfiguration
from gefyra import api

logger = logging.getLogger(__name__)


def remove_all_clients():
    clients = api.list_client()
    for client in clients:
        api.delete_client(client.client_id, force=True, wait=True)


def remove_remainder_bridges(config: ClientConfiguration):
    try:
        gbridges = config.K8S_CUSTOM_OBJECT_API.list_namespaced_custom_object(
            group="gefyra.dev",
            version="v1",
            namespace=config.NAMESPACE,
            plural="gefyrabridges",
        )
    except Exception:
        return None
    for bridge in gbridges.get("items"):
        try:
            config.K8S_CUSTOM_OBJECT_API.patch_namespaced_custom_object(
                group="gefyra.dev",
                version="v1",
                plural="gefyrabridges",
                namespace=config.NAMESPACE,
                name=bridge["metadata"]["name"],
                body={"metadata": {"finalizers": None}},
            )
            config.K8S_CUSTOM_OBJECT_API.delete_namespaced_custom_object(
                group="gefyra.dev",
                version="v1",
                plural="gefyrabridges",
                namespace=config.NAMESPACE,
                name=bridge["metadata"]["name"],
            )
        except Exception:
            continue
    return None


def remove_gefyra_namespace(config: ClientConfiguration):
    import kubernetes

    try:
        config.K8S_CORE_API.delete_namespace(name=config.NAMESPACE)
    except kubernetes.client.exceptions.ApiException as e:  # type: ignore
        if e.status == 404:
            return
        else:
            raise e from None


def remove_gefyra_crds(config: ClientConfiguration):
    import kubernetes

    crds = ["gefyrabridges.gefyra.dev", "gefyraclients.gefyra.dev"]
    for crd in crds:
        try:
            config.K8S_EXTENSION_API.delete_custom_resource_definition(name=crd)
        except kubernetes.client.exceptions.ApiException:  # type: ignore
            pass


def remove_gefyra_rbac(config: ClientConfiguration):
    import kubernetes

    clusterroles = ["gefyra:operator"]
    clusterrolebindings = ["gefyra-operator"]
    webhooks = ["gefyra.dev"]
    for cr in clusterroles:
        try:
            config.K8S_RBAC_API.delete_cluster_role(name=cr)
        except kubernetes.client.exceptions.ApiException as e:  # type: ignore
            logger.debug(e)
    for crb in clusterrolebindings:
        try:
            config.K8S_RBAC_API.delete_cluster_role_binding(name=crb)
        except kubernetes.client.exceptions.ApiException as e:  # type: ignore
            logger.debug(e)
    for wh in webhooks:
        try:
            config.K8S_ADMISSION_API.delete_validating_webhook_configuration(name=wh)
        except kubernetes.client.exceptions.ApiException as e:
            logger.debug(e)



import ast
import click
from gefyra.cli.utils import OptionEatAll, check_connection_name, parse_ip_port_map


@click.command()
@click.option(
    "-d",
    "--detach",
    help="Run container in background and print container ID",
    type=bool,
    default=False,
    is_flag=True,
)
@click.option(
    "--rm",
    "auto_remove",
    help="Automatically remove the container when it exits",
    type=bool,
    is_flag=True,
    default=False,
)
@click.option(
    "-p",
    "--expose",
    help="Add port mapping in form of <container_port>:<host_port>",
    type=str,
    multiple=True,
    callback=parse_ip_port_map,
)
@click.option(
    "--env-from",
    help="Copy the environment from the container in the notation 'Pod/Container'",
    type=str,
)
@click.option(
    "-v",
    "--volume",
    help=(
        "Bind mount a volume into the container in notation src:dest, allowed multiple"
        " times"
    ),
    type=str,
    multiple=True,
)
@click.option(
    "--env",
    help=(
        "Set or override environment variables in the form ENV=value, allowed multiple"
        " times"
    ),
    type=str,
    multiple=True,
)
@click.option(
    "-n", "--namespace", help="The namespace for this container to run in", type=str
)
@click.option(
    "-c",
    "--command",
    help="The command for this container to in Gefyra",
    type=str,
    cls=OptionEatAll,
)
@click.option(
    "-N",
    "--name",
    help="The name of the container running in Gefyra",
    type=str,
    required=True,
)
@click.option(
    "-i", "--image", help="The docker image to run in Gefyra", type=str, required=True
)
@click.option(
    "--connection-name", type=str, callback=check_connection_name, required=False
)
def run(
    detach,
    auto_remove,
    expose,
    env_from,
    volume,
    env,
    namespace,
    command,
    name,
    image,
    connection_name,
):
    from gefyra import api

    if command:
        command = ast.literal_eval(command)[0]
    api.run(
        image=image,
        name=name,
        command=command,
        namespace=namespace,
        env_from=env_from,
        env=env,
        ports=expose,
        auto_remove=auto_remove,
        volumes=volume,
        detach=detach,
        connection_name=connection_name,
    )



import click

from gefyra.cli.version import version
from gefyra.cli.run import run
from gefyra.cli.bridge import create_bridge, unbridge
from gefyra.cli.clients import clients
from gefyra.cli.connections import connections
from gefyra.cli.installation import install, uninstall
from gefyra.cli.status import status_command
from gefyra.cli.telemetry import telemetry
from gefyra.cli.updown import cluster_down, cluster_up
from gefyra.cli.list import list
from gefyra.cli.utils import AliasedGroup


@click.group(
    cls=AliasedGroup,
)
@click.option(
    "--kubeconfig",
    help="Path to the kubeconfig file to use instead of loading the default",
)
@click.option(
    "--context",
    help="Context of the kubeconfig file to use instead of 'default'",
)
@click.option("-d", "--debug", default=False, is_flag=True)
@click.pass_context
def cli(ctx: click.Context, kubeconfig, context, debug):
    import logging
    from gefyra.cli.telemetry import CliTelemetry

    if debug:
        logger = logging.getLogger()
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(levelname)s %(name)s %(filename)s:%(lineno)d - %(message)s"
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logging.getLogger("gefyra").setLevel(logging.DEBUG)
    else:
        logging.getLogger("gefyra").setLevel(logging.ERROR)
    ctx.ensure_object(dict)

    try:
        ctx.obj["telemetry"] = CliTelemetry()
    except Exception:  # pragma: no cover
        ctx.obj["telemetry"] = False
    ctx.obj["kubeconfig"] = kubeconfig
    ctx.obj["context"] = context


cli.add_command(cmd=telemetry, name="telemetry")
cli.add_command(cmd=connections, name="connections")
cli.add_command(cmd=clients, name="clients")
cli.add_command(cmd=install, name="install")
cli.add_command(cmd=uninstall, name="uninstall")
cli.add_command(cmd=cluster_up, name="up")
cli.add_command(cmd=cluster_down, name="down")
cli.add_command(cmd=status_command, name="status")
cli.add_command(cmd=create_bridge, name="bridge")
cli.add_command(cmd=unbridge, name="unbridge")
cli.add_command(cmd=run, name="run")
cli.add_command(cmd=version, name="version")
cli.add_command(cmd=list, name="list")


def main():
    cli(obj={})


if __name__ == "__main__":
    main()



from typing import Optional, List, TYPE_CHECKING
from gefyra.misc.comps import COMPONENTS
from gefyra.misc.utils import str_presenter

if TYPE_CHECKING:
    from gefyra.types import GefyraInstallOptions


def synthesize_config_as_dict(
    options: "GefyraInstallOptions", components: Optional[List[str]] = []  # noqa: B006
) -> list[dict]:
    req_comps = []
    if components:
        _comp_names = []
        for comp in COMPONENTS:
            comp_name = comp.__name__.split(".")[-1]
            _comp_names.append(comp_name)
            if comp_name in components:
                req_comps = [comp]
        if not req_comps or len(req_comps) != len(components):
            raise RuntimeError(
                f"Component(s) {','.join(components)} not found. "
                f"Choices are: {','.join(_comp_names)}"
            )
    else:
        req_comps = COMPONENTS

    data = []
    # generate the raw data
    for comp in req_comps:
        data.extend(comp.data(options))
    return data


def synthesize_config_as_yaml(
    options: "GefyraInstallOptions", components: Optional[List[str]] = []  # noqa: B006
) -> str:
    import yaml

    yaml.add_representer(str, str_presenter)
    data = synthesize_config_as_dict(options, components)
    _docs = []
    for doc in data:
        _docs.append(yaml.dump(doc))
    yaml_doc = ""
    if len(_docs) > 1:
        for idx, c in enumerate(_docs):
            if idx == 0:
                yaml_doc = c
            else:
                yaml_doc += f"\n---\n{c}"
    else:
        yaml_doc = _docs[0]

    return yaml_doc



import logging
from gefyra.configuration import ClientConfiguration
from gefyra.exceptions import GefyraConnectionError
from gefyra.types import StowawayConfig

logger = logging.getLogger("gefyra.cargo")


def get_cargo_ip_from_netaddress(network_address: str) -> str:
    return ".".join(network_address.split(".")[:3]) + ".149"


def probe_wireguard_connection(config: ClientConfiguration):
    from docker.models.containers import Container

    cargo: Container = config.DOCKER.containers.get(f"{config.CARGO_CONTAINER_NAME}")
    for _attempt in range(0, config.CARGO_PROBE_TIMEOUT * 2):
        logger.debug(
            f"Probing connection to {config.STOWAWAY_IP} (attempt {_attempt}/{config.CARGO_PROBE_TIMEOUT * 2}))"
        )
        _r = cargo.exec_run(f"timeout 1 ping -c 1 {config.STOWAWAY_IP}")
        if _r.exit_code != 0:
            continue
        else:
            break
    else:
        raise GefyraConnectionError(
            "Gefyra could not successfully confirm the connection working."
        )


def create_wireguard_config(
    params: StowawayConfig, cargo_endpoint: str, mtu: str = "1340"
) -> str:
    return (
        "[Interface]\n"
        f"Address = {params.iaddress}\n"
        f"MTU = {mtu}\n"
        f"PrivateKey = {params.iprivatekey}\n"
        f"DNS = {params.idns}\n"
        "PreUp = sysctl -w net.ipv4.ip_forward=1\n"
        "PostUp = iptables -A FORWARD -i %i -j ACCEPT; iptables -A FORWARD -o %i -j ACCEPT; iptables -t nat -A POSTROUTING -o eth1 -j MASQUERADE\n"  # noqa E501
        "PostDown = iptables -D FORWARD -i %i -j ACCEPT; iptables -D FORWARD -o %i -j ACCEPT; iptables -t nat -D POSTROUTING -o eth1 -j MASQUERADE\n"  # noqa E501
        "[Peer]\n"
        f"PublicKey = {params.ppublickey}\n"
        f"Endpoint = {cargo_endpoint}\n"
        f"PresharedKey = {params.presharedkey}\n"
        "PersistentKeepalive = 21\n"
        "AllowedIPs = 0.0.0.0/0\n"
    )



# flake8: noqa
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from gefyra.types import GefyraInstallOptions


def data(params: "GefyraInstallOptions") -> list[dict]:
    return [
        {
            "apiVersion": "v1",
            "kind": "ServiceAccount",
            "metadata": {"namespace": params.namespace, "name": "gefyra-operator"},
        },
        {
            "apiVersion": "rbac.authorization.k8s.io/v1",
            "kind": "ClusterRole",
            "metadata": {"name": "gefyra:operator"},
            "rules": [
                {
                    "apiGroups": ["kopf.dev"],
                    "resources": ["clusterkopfpeerings"],
                    "verbs": ["list", "watch", "patch", "get"],
                },
                {
                    "apiGroups": ["apiextensions.k8s.io"],
                    "resources": ["customresourcedefinitions"],
                    "verbs": ["create", "patch", "delete", "list", "watch"],
                },
                {
                    "apiGroups": ["admissionregistration.k8s.io"],
                    "resources": [
                        "validatingwebhookconfigurations",
                        "mutatingwebhookconfigurations",
                    ],
                    "verbs": ["create", "patch"],
                },
                {
                    "apiGroups": [
                        "",
                        "apps",
                        "batch",
                        "extensions",
                        "events.k8s.io",
                        "rbac.authorization.k8s.io",
                        "rbac.authorization.k8s.io/v1",
                    ],
                    "resources": [
                        "namespaces",
                        "clusterroles",
                        "serviceaccounts",
                        "clusterrolebindings",
                        "nodes",
                        "configmaps",
                        "secrets",
                        "deployments",
                        "statefulsets",
                        "persistentvolumeclaims",
                        "services",
                        "pods",
                        "pods/exec",
                        "events",
                    ],
                    "verbs": ["*"],
                },
                {
                    "apiGroups": ["gefyra.dev"],
                    "resources": ["gefyraclients", "gefyrabridges"],
                    "verbs": ["*"],
                },
            ],
        },
        {
            "apiVersion": "rbac.authorization.k8s.io/v1",
            "kind": "ClusterRoleBinding",
            "metadata": {
                "name": "gefyra-operator",
                "namespace": params.namespace,
            },
            "roleRef": {
                "apiGroup": "rbac.authorization.k8s.io",
                "kind": "ClusterRole",
                "name": "gefyra:operator",
            },
            "subjects": [
                {
                    "kind": "ServiceAccount",
                    "name": "gefyra-operator",
                    "namespace": params.namespace,
                }
            ],
        },
    ]



import logging
from pathlib import Path
from typing import Iterable, List, Optional
import uuid
from gefyra.configuration import ClientConfiguration
from gefyra.local.clients import (
    get_gefyraclient_body,
    handle_create_gefyraclient,
    handle_delete_gefyraclient,
    handle_get_gefyraclient,
)
from gefyra.types import GefyraClient
from .utils import stopwatch

logger = logging.getLogger(__name__)


@stopwatch
def add_clients(
    client_id: str,
    quantity: int = 1,
    kubeconfig: Optional[Path] = None,
    kubecontext: Optional[str] = None,
) -> Iterable[GefyraClient]:
    """
    Add a new client to the connection provider
    """
    config = ClientConfiguration(
        kube_config_file=kubeconfig, kube_context=kubecontext, ignore_connection=True
    )
    if quantity > 1 and client_id:
        raise RuntimeError("Cannot specify both quantity > 1 and client_id")
    result: List[GefyraClient] = []
    while len(result) < quantity:
        if not bool(client_id):
            generated_uuid = uuid.uuid4()
            client_id = str(generated_uuid).replace("-", "")

        logger.info(f"Creating client with id: {client_id}")
        gclient_req = get_gefyraclient_body(config, client_id)
        gclient = handle_create_gefyraclient(config, gclient_req)
        result.append(GefyraClient(gclient, config))
        client_id = ""
    return result


@stopwatch
def get_client(
    client_id: str,
    connection_name: str = "",
    kubeconfig: Optional[Path] = None,
    kubecontext: Optional[str] = None,
) -> GefyraClient:
    """
    Get a GefyraClient object
    """
    config_params = {"connection_name": connection_name}
    if kubeconfig:
        config_params.update({"kube_config_file": str(kubeconfig)})

    if kubecontext:
        config_params.update({"kube_context": kubecontext})
    config = ClientConfiguration(**config_params)  # type: ignore
    gclient = handle_get_gefyraclient(config, client_id)
    return GefyraClient(gclient, config)


@stopwatch
def delete_client(
    client_id: str,
    force: bool = False,
    kubeconfig: Optional[Path] = None,
    kubecontext: Optional[str] = None,
    connection_name: Optional[str] = None,
    wait: bool = False,
) -> bool:
    """
    Delete a GefyraClient configuration
    """
    config = ClientConfiguration(
        kube_config_file=kubeconfig,
        kube_context=kubecontext,
        connection_name=connection_name if connection_name else "no-connection-name",
        # use no-connection-name to make sure you use admin access to the cluster
    )
    return handle_delete_gefyraclient(config, client_id, force, wait=wait)


@stopwatch
def write_client_file(
    client_id: str,
    host: Optional[str] = None,
    port: Optional[str] = None,
    kube_api: Optional[str] = None,
    kubeconfig: Optional[Path] = None,
    kubecontext: Optional[str] = None,
) -> str:
    """
    Write a client file
    """
    config = ClientConfiguration(kube_config_file=kubeconfig, kube_context=kubecontext)
    client = get_client(
        client_id, kubeconfig=config.KUBE_CONFIG_FILE, kubecontext=config.KUBE_CONTEXT
    )
    if not port:
        port = "31820"
    if host:
        gefyra_server = f"{host}:{port}"
    else:
        gefyra_server = config.get_stowaway_host(port)
    logger.debug(f"gefyra_server: {gefyra_server}")
    return client.get_client_config(
        gefyra_server=gefyra_server, k8s_server=kube_api
    ).json


@stopwatch
def list_client(
    kubeconfig: Optional[Path] = None, kubecontext: Optional[str] = None
) -> List[GefyraClient]:
    """
    List all GefyraClient objects
    """
    config = ClientConfiguration(kube_config_file=kubeconfig, kube_context=kubecontext)
    clients = config.K8S_CUSTOM_OBJECT_API.list_namespaced_custom_object(
        namespace=config.NAMESPACE,
        group="gefyra.dev",
        plural="gefyraclients",
        version="v1",
    )
    return [GefyraClient(client, config) for client in clients["items"]]
